#!/usr/local/bin/perl -w
use strict;
use TIGR::Foundation;
use File::Path;

my $USAGE = "Usage: addCoverage.pl reference.slice reads.spec reads.snps [OPTIONS]\n";
my $HELPTEXT = qq~
Adds multiple reads onto a preexisting assembly. The preexisting assembly
should provided in slice format (reference.slice). The reads to add should
all be provided in slice format, and should all be available within a
directory specified by <seqdir>. The reads to merge should be listed in the
spec file, which lists the seqname and alignment chain to use for merging 
(one sequence per line). The snps file will be used to align the reads using
the chain specified in the spec file. The program will error if you try to
add two reads to the reference at the same time if they overlap.

  $USAGE
    reference.slice Reference Assembly to add reads to
    reads.spec      File containing (seqname chain) pairs to add
    reads.snps      File containing snp data for all reads to add

  options:
    -s <seqdir> Specify directory that contains the read slice files
                                           (DEFAULT: seqs)

    -o <prefix>  Specify prefix for result (DEFAULT: merge)
    -w <workdir> Specify work directory    (DEFAULT: work)

    -[no]merge  [Don't] merge the resultant slice file (DEFAULT: -merge)
    -[no]recall [Don't] Recall the merged consensus    (DEFAULT: -norecall)
    -[no]contig [Don't] Convert result to contig       (DEFAULT: -nocontig)
    -[no]tcov   [Don't] Convert result to tcov         (DEFAULT: -notcov)

    -restart    Attempt to use intermediate files from previous run. 
                Default: erase previous work directory and restart from scratch

    -0 Input slices are 0-based [Default]
    -1 Input slices are 1-based

    -v Toggle being more verbose
~;


my $VERSION = " Version 1.20 (Build " . (qw/$Revision: 1.1.1.1 $/ )[1] . ")";

my $GETCOVERAGE  = "%BINDIR%/getCoverage";
my $MERGESLICE   = "%BINDIR%/mergeSlice";
my $SLICE2CONTIG = "%BINDIR%/slice2contig";
my $SPLITSLICE   = "%BINDIR%/splitSlice";
my $ZIPCLAP      = "%BINDIR%/zipclap";
my $TRSLICE      = "%BINDIR%/trSlice";

my $GREP         = "/bin/grep";

my @DEPENDS = 
(
  "TIGR::Foundation",
  $GETCOVERAGE,  
  $MERGESLICE,   
  $SLICE2CONTIG, 
  $SPLITSLICE,   
  $ZIPCLAP,      
  $TRSLICE,
  $GREP
);


my $tf = new TIGR::Foundation;
my $doverbose = 0;

my $use1based = 0;

sub echo
{
  my $str = join " ", @_;

  $tf->logLocal($str, 4);
  print @_ if ($doverbose);
}

sub runCmd
{
  my $cmd = shift;
  my $info = shift;

  echo "$info... " if defined $info;

  my $exitcode = $tf->runCommand($cmd);

  if ($exitcode)
  {
    echo "failed! ($!)\n" if defined $info;
    $tf->bail("Error with $cmd ($exitcode)");
  }

  echo "ok.\n" if defined $info;
}

my $mergefile = "mergelist";
my $splitfile = "splitlist";

my %seqs;
my @seqlist;

sub splitReference
{
  my $REFERENCE = shift;

  symlink "../$REFERENCE.slice", "$REFERENCE.slice" if (! -e "$REFERENCE.slice");

  ## Convert from 1-based split (alignment) coordinate to 0-based slices (if necessary)
  my $shift = ($use1based) ? 0 : 1;
  my @splitlist = map {$seqs{$_}->{asm_l}-$shift} @seqlist;
  shift @splitlist; ## Don't split before the first element

  if (scalar @splitlist)
  {
    ## The reference needs to be split

    open SPLIT, "> $splitfile" 
      or tf->bail("Can't open $splitfile for writing ($!)");

    print SPLIT join("\n", @splitlist), "\n";
    close SPLIT;

    runCmd("$SPLITSLICE -f $splitfile $REFERENCE.slice")
      if (! -e $REFERENCE."_0.slice");
  }
  else
  {
    ## The reference does not need to be split, so just symlink for consistency

    if (! -e $REFERENCE."_0.slice")
    {
      symlink ("$REFERENCE.slice", "$REFERENCE\_0.slice")
        or $tf->bail("Couldn't symlink $REFERENCE.slice to $REFERENCE\_0.slice ($!)");
    }
  }
}

sub zipseq
{
  my $seqname = shift;
  my $piece = shift;
  my $snpfile = shift;
  my $REFERENCE = shift;
  my $SEQDIR = shift;

  my $sr = $seqs{$seqname};
  my $chain = $sr->{chain};

  ## Convert from 1-based alignment to 0-based shift 
  my $refshift = ($piece == 0) ? 0 : $sr->{asm_l}-1;

  ## Create a directory to work in
  mkdir $piece if (! -e $piece);
  chdir $piece;

  $tf->bail("Can't read ../../$SEQDIR/$seqname.slice")
    if (! -r "../../$SEQDIR/$seqname.slice");

  symlink "../../$SEQDIR/$seqname.slice", "$seqname.slice" if (! -r "$seqname.slice");
  symlink "../../$SEQDIR/$seqname.fasta", "$seqname.fasta" if (! -r "$seqname.fasta" && -r "../../$SEQDIR/$seqname.fasta");

  $tf->bail("Can't read ../$REFERENCE\_$piece.slice")
    if (! -r "../$REFERENCE\_$piece.slice");

  symlink "../$REFERENCE\_$piece.slice", "$REFERENCE.slice" if (! -r "$REFERENCE.slice");

  ## Extract the snp information for this sequence
  runCmd("$GREP \" $seqname,\" ../../$snpfile > zip.snps")
    if (! -e "zip.snps");

  if (! -e "zip.slice")
  {
    my $v = "";
    $v = "-v" if ($doverbose);

    runCmd("$ZIPCLAP $REFERENCE $seqname -c $chain -r $refshift -nocontig -q $seqname.fasta $v");
  }

  chdir "..";
}


sub mergePieces
{
  my $domerge = shift;
  my $docontig = shift;
  my $dotcov = shift;
  my $dorecall = shift;

  my $mergeprefix = shift;
  my $workdir = shift;

  if (! -e "$mergeprefix.slice")
  {
    open MERGE, "> $mergefile" 
      or $tf->bail("Can't open $mergefile for writing ($!)");

    foreach my $piece (@_) { print MERGE "$piece\n"; }
    close MERGE;

    if ($domerge)
    {
      print "Merging...\n";

      my $recall = ($dorecall) ? "-c -a 3" : "";
      my $trslice = "$TRSLICE $recall -x $use1based -y $use1based -i $use1based -g $use1based";

      if (scalar @_ > 1)
      {
        runCmd("$MERGESLICE -A $mergefile | $trslice - -o ../$mergeprefix.slice");
      }
      else
      {
        my $piece = $_[0];
        runCmd("$trslice $piece -o ../$mergeprefix.slice --debug 4");
      }

      chdir "..";

      if ($docontig)
      {
        runCmd("$SLICE2CONTIG -i $mergeprefix $mergeprefix.slice -o $mergeprefix.contig");

        runCmd("$GETCOVERAGE -t -M --silent --gapped $mergeprefix")
          if ($dotcov);
      }
    }
  }
}

sub verifyOverlaps
{
  my $listsize = (scalar @seqlist)-1;

  for (my $i = 0; $i < $listsize; $i++)
  {
    my $s1 = $seqlist[$i];
    my $asml1 = $seqs{$s1}->{asm_l};
    my $asmr1 = $seqs{$s1}->{asm_r};

    my $s2 = $seqlist[$i+1];
    my $asml2 = $seqs{$s2}->{asm_l};
    my $asmr2 = $seqs{$s2}->{asm_r};

    ## Verify there are no overlaps between the sequences
    if ($asmr1 >= $asml2)
    {
      $tf->bail("Overlap detected between $s1 <$asml1 $asmr1> ".
                "and $s2 <$asml2 $asmr2>\n");
    }
  }
}

sub loadSpec
{
  my $specfile = shift;

  open SPEC, "< $specfile" 
    or $tf->bail("Can't open $specfile ($!)");

  while (<SPEC>)
  {
    chomp;

    next if /^$/;

    my ($seqname, $chain) = split /\s+/, $_;
    if (!defined ($chain) || $chain =~ /\D/)
    {
      $tf->bail("Invalid chain specified for $seqname");
    }

    $seqs{$seqname}->{chain} = $chain;
  }
  close SPEC;
}

sub loadSNPs
{
  my $snpfile = shift;

  open SNPS, "< $snpfile" 
    or $tf->bail("Can't open $snpfile ($!)");

  while (<SNPS>)
  {
    chomp;

    my @fields = split /, /, $_;

    my $seqname = $fields[3];
    my $chain   = $fields[4];
    my $asm_l   = $fields[5];
    my $asm_r   = $fields[6];
    
    if (exists $seqs{$seqname})
    {
      if (($seqs{$seqname}->{chain} eq $chain) &&
          !defined $seqs{$seqname}->{asm_l})
      {
        $seqs{$seqname}->{asm_l} = $asm_l;
        $seqs{$seqname}->{asm_r} = $asm_r;
      }
    }
  }
  close SNPS;

  @seqlist = sort {$seqs{$a}->{asm_l} <=> $seqs{$b}->{asm_l}} keys %seqs;
}



MAIN:
{
  use Cwd;

  $tf->addDependInfo(@DEPENDS);
  $tf->setHelpInfo($HELPTEXT);
  $tf->setVersionInfo($VERSION);
  $tf->setLogFile(getcwd()."/addCoverage.log"); ## Don't follow the chdir

  my $domerge = 1;
  my $docontig = 0;
  my $dotcov = 0;
  my $dorecall = 0;

  my $seqdir = "seqs";
  my $mergeprefix = "merge";
  my $workdir = "work";

  my $do0based = 0;
  my $do1based = 0;

  my $tryrestart = 0;

  # now we handle the input options
  my $result = $tf->TIGR_GetOptions
               (
                 's|seqdir=s', \$seqdir,
                 'o=s',        \$mergeprefix,
                 'w=s',        \$workdir,
                 'merge!',     \$domerge,
                 'contig!',    \$docontig,
                 'tcov!',      \$dotcov,
                 'recall!',    \$dorecall,
                 'restart',    \$tryrestart,
                 '0',          \$do0based,
                 '1',          \$do1based,

                 'v|verbose!',  \$doverbose,
               );

  $tf->bail("Command line parsing failed") if (!$result);

  if ($do0based) { $use1based = 0; }
  else           { $use1based = 1; }

  my $reference = shift || die $USAGE;
  my $specfile  = shift || die $USAGE;
  my $snpfile   = shift || die $USAGE;

  my $debug = $tf->getDebugLevel();
  $tf->setDebugLevel(4) if (!$debug);

  $reference =~ s/\.slice$//;

  $docontig = 1 if ($dotcov);

  print "Loading Spec...\n";
  loadSpec($specfile);

  print "Loading SNPs...\n";
  loadSNPs($snpfile);

  print "Verifying Overlaps...\n";
  verifyOverlaps();

  if (-e $workdir && !$tryrestart)
  {
    rmtree ($workdir) 
      or $tf->bail("Can't cleanup old workdir $workdir ($!)\n");
  }

  if (! -e $workdir)
  {
    mkdir $workdir
      or $tf->bail("Can't create workdir $workdir\n");
  }

  chdir $workdir
    or $tf->bail("Can't chdir to workdir $workdir\n");

  print "Splitting...\n";
  splitReference($reference);

  my @mergelist;

  print "Zipping...\n";
  my $i = 0;
  foreach my $seqname (@seqlist)
  {
    my $asml = $seqs{$seqname}->{asm_l};
    my $asmr = $seqs{$seqname}->{asm_r};

    print "$i $seqname <$asml $asmr>\n";
    zipseq($seqname, $i, $snpfile, $reference, $seqdir);

    push @mergelist, "$i/zip.slice";

    $i++;
  }

  print "Collecting Pieces... \n";
  mergePieces($domerge, $docontig, $dotcov, $dorecall, $mergeprefix, $workdir, @mergelist);
}



