#include "Contig.hh"
Contig::Contig(long index, 
               long gindex, 
               ReadManager * readManager,
               bool verbose)
 : m_tiling(m_readVector, m_gindex)
{
  m_verbose = verbose;
  m_readManager = readManager;
  m_newIndex = 0;
  m_newGindex = 0;
  m_gindex = 0;
  m_readCount = 0;

  m_contigName = "0";

  m_deleteReadManager = false;

  m_sliceRangeIndex = index;
  m_sliceRangeGindex = gindex;
}

Contig::~Contig()
{
  ReadVector::iterator r;

  for (r =  m_readVector.begin();
       r != m_readVector.end();
       r++)
  {
    delete *r;
  }

  if (m_deleteReadManager)
  {
    delete m_readManager;
  }
}

ReadVector & Contig::getReadVector()
{
  return m_readVector;
}

long Contig::getWidth()
{
  return m_gindex;
}

void Contig::loadContig(ifstream * cf)
{
  m_deleteReadManager = true;
  m_readManager = new ReadManager();

  string contigname;
  int numseqs;
  int width;
  (*cf) >> contigname;
  int currentSeq = 0;

  if (contigname[0] == contigname[1] && contigname[1] == '#')
  {
    //load the contig header 
    contigname.erase(0,2);

    m_contigName = contigname;

    (*cf) >> numseqs;
    (*cf) >> width;

    m_gindex = width+1;

    // Drop the checksum
    char buffer[1024*1024];
    cf->getline(buffer, sizeof(buffer));

    // Load the consensus
    while ((*cf) && (cf->peek() != '#'))
    {
      cf->getline(buffer, sizeof(buffer));
      m_consensus += buffer;
    }


    while (currentSeq < numseqs)
    {
      // load the header line
      cf->getline(buffer, sizeof(buffer));

      string seqname;
      int offset;
      int pos = 1; // skip the '#'

      // Copy the seqname
      while (buffer[pos] != '(')
      {
        seqname += buffer[pos];
        pos++;
      }

      pos++; // skip the (
      
      //read the offset
      offset = atoi(buffer+pos);

      while (buffer[pos] != '[') { pos++; }
      pos++; 

      int rc = (buffer[pos] != ']');
      while (buffer[pos] != ']') { pos++; }
      pos++;

      //int len = atoi(buffer+pos);

      while(buffer[pos] != '{') { pos++; }
      pos++;


      string seq_l;
      while(buffer[pos] != ' ') 
      { 
        seq_l += buffer[pos];
        pos++; 
      }

      pos++;

      string seq_r;
      while(buffer[pos] != '}') 
      { 
        seq_r += buffer[pos];
        pos++; 
      }

      string lsequence;

      // Load the lsequence
      while ((*cf) && (cf->peek() != '#'))
      {
        cf->getline(buffer, sizeof(buffer));
        lsequence += buffer;
      }

      sprintf(buffer, "%d", currentSeq);

      // Add the read to the read manager
      ReadData * rd = m_readManager->addReadData(buffer,
                                                 seqname.c_str(),
                                                 rc ? "1" : "0", "");
      rd->addClrData(seq_l.c_str(), seq_r.c_str());

      // Now add it to the read map
      ContigSequence * cs = newContigSequence(rd, offset);

      // Load the sequence into it
      cs->addSequence(lsequence, "");

      currentSeq++;
    }
  }
}


void Contig::addReadToTiling(ContigSequence * read, 
                             long startGindex, 
                             int numPositions)
{
  m_tiling.setDirty(startGindex, numPositions);
}


void Contig::addSlice(SliceData * slice)
{
  unsigned int depth = slice->getDepth();
  char cons = slice->getConsensus();

  if (depth)
  {
    // Recompute an unambiguous consensus based on cumulative qv
    int scores[5] = {0,0,0,0,0};
    char cbases[]  = "ACGT-";
    const string & bases = slice->getNucData();

    for (unsigned int i = 0; i < depth; i++)
    {
      ReadData * rd = slice->getRead(i);
      ContigSequence * read = getContigSequence(rd->getSeqname());

      if (read == NULL)
      {
        read = newContigSequence(rd, m_gindex);
      }

      char base = bases[i];
      int qv = slice->getQualVal(i);
      read->addSlice(base, qv);

      switch(base)
      {
        case 'A': case 'a': scores[0] += 1; break;
        case 'C': case 'c': scores[1] += 1; break;
        case 'G': case 'g': scores[2] += 1; break;
        case 'T': case 't': scores[3] += 1; break;
        case '-':           scores[4] += 1; break;
      }
    }

    // Find the base with the greatest score
    int consc = 0;
    for (int i = 1; i < 5; i++)
    {
      if (scores[i] > scores[consc]) { consc = i; }
    }

    cons = cbases[consc];
  }

// Can't call this because it uses the tiling
//  calculateCQV(m_gindex, cons);
  m_consensus += cons;

  m_gindex++;

  delete slice;
}



void Contig::stripSequenceEdgeGaps()
{
  ReadVector::iterator r;

  for (r =  m_readVector.begin();
       r != m_readVector.end();
       r++)
  {
    ContigSequence * seq = *r;

    // Strip the left gaps
    long origoffset = seq->getOffset();
    long newoffset = seq->stripLeftGaps();

    if (newoffset != origoffset)
    {
      m_tiling.setDirty(origoffset, newoffset-origoffset);
    }

    // String the right gaps
    long origend = seq->getRightEnd();
    long newend = seq->stripRightGaps();
    
    if (newend != origend)
    {
      m_tiling.setDirty(newend+1, origend-newend);
    }
  }
}

long Contig::shiftContigRight(int contigShiftDistance)
{
  cerr << "@ContigShiftDistance=" << contigShiftDistance << endl;

  // All of the downstream offsets need to be adjusted
  ReadVector::iterator r;
  for (r =  m_readVector.begin();
       r != m_readVector.end();
       r++)
  {
    (*r)->adjustOffset(0, contigShiftDistance);
  }

  // open space in the consensus and tiling
  m_consensus.insert(m_consensus.begin(), contigShiftDistance, '-');
  m_tiling.shiftTilingRight(contigShiftDistance);
  m_gindex += contigShiftDistance;

  return m_gindex;
}

long Contig::widenContig(int contigGrowDistance)
{
  cerr << "@ContigGrowDistance=" << contigGrowDistance << endl;

  m_consensus.append(contigGrowDistance, '-');
  m_tiling.resizeTiling(contigGrowDistance);
  m_gindex += contigGrowDistance;

  return m_gindex;
}

long Contig::trimContigRight(int newContigSize)
{
  int distance = m_gindex - newContigSize;
  cerr << "@TrimContigRight=" << distance << endl;

  m_consensus.resize(newContigSize);
  m_tiling.resizeTiling(newContigSize - m_gindex);
  m_gindex = newContigSize;

  return m_gindex;
}

long Contig::trimContigLeft(int trimBases)
{
  cerr << "@TrimContigLeft=" << trimBases << endl;

  // All of the downstream offsets need to be adjusted
  ReadVector::iterator r;
  for (r =  m_readVector.begin();
       r != m_readVector.end();
       r++)
  {
    (*r)->adjustOffset(0, -trimBases);
  }

  m_consensus.erase(0, trimBases);
  m_tiling.trimTilingLeft(trimBases);
  m_gindex -= trimBases;

  return m_gindex;
}

ContigSequence * Contig::newContigSequence(ReadData * rd, long gindex)
{
  ContigSequence * retval = new ContigSequence(rd, gindex, m_readCount);
  addContigSequence(retval);
  return retval;
}

void Contig::addContigSequence(ContigSequence * read)
{
  m_readVector.push_back(read);
  m_readMap[read->getSeqname()] = m_readCount;
  m_readCount++;
}

ContigSequence * Contig::getContigSequence(const string & seqname)
{
  ContigSequence * retval = NULL;

  ReadMap::const_iterator i = m_readMap.find(seqname);
  if (i != m_readMap.end())
  {
    retval = m_readVector[i->second];
  }

  return retval;
}

ContigSequenceSlice & Contig::getTilingSlice(long gindex)
{
  return m_tiling.getTilingSlice(gindex);
}

void Contig::stripGapSlices()
{
  for (long gindex = m_gindex-1; gindex >= 0; gindex--)
  {
    const ContigSequenceSlice & tilingSlice = getTilingSlice(gindex);

    ContigSequenceSlice::const_iterator t;
    unsigned int depth = tilingSlice.size();
    bool allGap = depth; // never all gap in zero coverage

    for (t =  tilingSlice.begin();
         t != tilingSlice.end();
         t++)
    {
      if ((*t)->base(gindex) != '-')
      {
        allGap = false;
        break;
      }
    }

    if (allGap)
    {
      if (m_verbose) { cerr << "All gap at " << gindex << endl; }
      
      // remove the entire column of tiling, 
      // effectively slide the tiling one position left
      m_consensus.erase(gindex, 1);
      m_tiling.eraseTilingSlice(gindex);
      m_gindex--;

      // Erase the gap (and shift downstream offsets)
      ReadVector::iterator r;
      for (r =  m_readVector.begin();
           r != m_readVector.end();
           r++)
      {
        (*r)->erase(gindex);
      }
    }
  }
}

void Contig::printPos(long gindexstart, long gindexend)
{
  long maxvalue = gindexend;
  long divide = 1;
  int numDigits = 1;

  while (maxvalue > 9)
  {
    divide *= 10;
    maxvalue /= 10;
    numDigits++;
  }

  for (; numDigits > 0; numDigits--)
  {
    cerr << "pos:\t";
    for (long j = gindexstart; j <= gindexend; j++)
    {
      int currentValue = (j/divide) % 10;
      cerr << currentValue;
    }
    cerr << endl;

    divide /= 10;
  }
}



void Contig::dumpRange(long gindexstart, 
                       long gindexend, 
                       ContigSequence * read)
{
  if (!m_verbose) { return; }

  if (gindexstart < 0)       { gindexstart = 0;}
  if (gindexend >= m_gindex) { gindexend = m_gindex-1; }

  printPos(gindexstart, gindexend);

  cerr << "cons:\t" 
       << m_consensus.substr(gindexstart, gindexend - gindexstart +1)
       << endl;

  if (read)
  {
    read->dumpRange(gindexstart, gindexend);
  }
  else
  {
    ReadVector::iterator r;
    for (r =  m_readVector.begin();
         r != m_readVector.end();
         r++)
    {
      (*r)->dumpRange(gindexstart, gindexend);
    }
  }
}

void Contig::dumpAlignment()
{
  if (!m_verbose) { return; }

  printPos(0, m_gindex-1);

  cerr << "cons:\t" <<  m_consensus << endl;
  ReadVector::iterator r;
  for (r =  m_readVector.begin();
       r != m_readVector.end();
       r++)
  {
    (*r)->dump();
  }
}



void Contig::recallCons()
{
  for (int gindex = 0; gindex < m_gindex; gindex++)
  {
    char cons = m_consensus[gindex];
    calculateCQV(gindex, cons);
    m_consensus[gindex] = cons;
  }
}

int Contig::calculateCQV(long gindex, char & cons)
{
  // Store the count of ACGT-
  char cbases[]  = "ACGT-";
  int scores[5] = {0,0,0,0,0};

  ContigSequenceSlice::iterator end;
  ContigSequenceSlice::iterator t;

  m_tiling.setTilingIterators(gindex, t, end);

  if (t == end)
  {
    // Don't recall the consensus of empty slices
    return 0;
  }

  for (; t != end; t++)
  {
    ContigSequence * currentRead = *t;
    switch(currentRead->base(gindex))
    {
      case 'A': case 'a': scores[0] += 1; break;
      case 'C': case 'c': scores[1] += 1; break;
      case 'G': case 'g': scores[2] += 1; break;
      case 'T': case 't': scores[3] += 1; break;
      case '-':           scores[4] += 1; break;
    }
  }

  unsigned int i;
  unsigned int max = 0; // Assume A is max
  int maxcount = scores[max];
  for (i = 1; i < 5; i++)
  {
    if (scores[i] > maxcount)
    {
      max = i;
      maxcount = scores[i];
    }
  }

  cons = cbases[max];

  // Get rough cqv by doubling counts, subtracting the conflicting counts, 
  // but prefer gap misalignment to base misalignment
  maxcount *= 2;
  
  for (i = 0; i < 5; i++)
  {
    if (i != max)
    {
      maxcount -= (i == 4) ? scores[i] : 2*scores[i];
    }
  }

  return maxcount;
}



long Contig::insertGapSlice(long gindex, ContigSequence * read)
{
  if (m_verbose) { cerr << "inserting gap slice at " << gindex << endl; }
  m_consensus.insert(gindex, "-");

  m_tiling.insertGapSlice(gindex);
  m_gindex++;

  // shift the offset of the downstream reads
  ReadVector::iterator r;
  for (r =  m_readVector.begin();
       r != m_readVector.end();
       r++)
  {
    if (*r == read)
    {
      (*r)->adjustOffset(gindex, +1);
    }
    else
    {
      (*r)->insertGap(gindex);
    }
  }

  return m_gindex;
}

long Contig::getUngappedIndex(long int gappedIndex)
{
  long retval = 1;
  long pos = 0;

  while (gappedIndex)
  {
    if (m_consensus[pos] != '-')
    {
      retval++;
    }

    gappedIndex--;
    pos++;
  }

  return retval;
}

string Contig::getUngappedConsensus(long int startindex,
                                    long int length)
{
  // Strip the gaps from the consensus
  string cons = m_consensus.substr(startindex, length);
  string::size_type erasepos = 0;

  while ((erasepos = cons.find('-', erasepos)) != cons.npos)
  {
    cons.erase(cons.begin()+erasepos);
  }

  return cons;
}

long Contig::getGappedIndex(long int ungappedIndex)
{
  long int gindex = 0;

  while (ungappedIndex)
  {
    if (m_consensus[gindex] != '-')
    {
      ungappedIndex--;
    }
    gindex++;
  }

  while (m_consensus[gindex] == '-')
  {
    gindex++;
  }

  return gindex;
}


void Contig::printContig(ostream & out, PrintType displayType)
{
  if (displayType == CONTIG)
  {
    out << "##" << m_contigName
        << " "  << m_readVector.size() 
        << " "  << m_consensus.length() 
        << " bases, 00000000 checksum." << endl;

    RequestManager::print60(out, m_consensus);
    ReadVector::iterator i;
    for (i =  m_readVector.begin();
         i != m_readVector.end();
         i++)
    {
      long asm_l = getUngappedIndex((*i)->getOffset());
      long asm_r = getUngappedIndex((*i)->getRightEnd());


      (*i)->printContigSequence(out, asm_l, asm_r);
    }
  }
  else
  {
    if (displayType == SLICE)
    {
      out << "  <ReadCollection>" << endl;

      ReadVector::iterator i;
      for (i =  m_readVector.begin();
           i != m_readVector.end();
           i++)
      {
        (*i)->printReadData(out);
      }

      out << "  </ReadCollection>" << endl;
      out << "  <SliceRange Index=\"" << m_sliceRangeIndex
          << "\" Gindex=\""           << m_sliceRangeGindex
          << "\">" << endl;
    }

    m_newIndex = 0;
    m_newGindex = 0;

    if (displayType == SLICE || displayType == TCOV)
    {
      const char z[] = "0";

      SliceData * currentSlice = new SliceData(*m_readManager, z, z, z);

      for (long gindex = 0; gindex < m_gindex; gindex++)
      {
        char consensus = m_consensus[gindex];

        currentSlice->resetNuc();
        currentSlice->resetQualVal();
        currentSlice->resetReadId();
        currentSlice->setConsensus(consensus);

        //cerr << gindex << ":" << consensus << endl;

        ContigSequenceSlice & tilingSlice = m_tiling.getTilingSlice(gindex);
        ContigSequenceSlice::const_iterator t;

        try
        {
          tilingSlice.sort(ContigSeqCmp());

          for (t =  tilingSlice.begin();
               t != tilingSlice.end();
               t++)
          {
            ContigSequence * currentContigSequence = *t;
            currentSlice->addNuc(currentContigSequence->getNuc(gindex));

            if (t != tilingSlice.begin())
            {
              currentSlice->addQualVal(" ");
              currentSlice->addReadId(" ");
            }

            currentSlice->addQualVal(currentContigSequence->getQual(gindex));
            currentSlice->addReadId(currentContigSequence->getReadId());
          }

          long coordinate = m_newIndex;
          if (coordinate && consensus == '-') coordinate--;

          currentSlice->resetIndex(coordinate, m_newGindex);

          switch (displayType)
          {
            case SLICE: printSlice(out, currentSlice); break;
            case TCOV:  printSliceTcov(out, currentSlice); break;
            case CONTIG: break;
          }

          if (consensus != '-') m_newIndex++;
          m_newGindex++;
        }
        catch (const Exception & e)
        {
          cerr << e << endl;

          cerr << "Tiling at " << gindex << ":" << endl;
          for (t =  tilingSlice.begin();
               t != tilingSlice.end();
               t++)
          {
            cerr << " " << (*t)->getReadCount();
          }
          cerr << endl;

          throw e;
        }
      }

      if (displayType == SLICE)
      {
        out << "  </SliceRange>" << endl;
      }

      delete(currentSlice);
    }
  }
}


void Contig::printSlice(ostream & out, SliceData * slice)
{
  //slice->setDoAmbiguity(0);
  //slice->doCalculations(); 

  slice->printXML(out);
}

string & Contig::getConsensus()
{
  return m_consensus;
}

void Contig::printSliceTcov(ostream & out, SliceData * slice)
{
  out << slice->getGindex() << " "
      << slice->getIndex()  << " ";

  slice->setDoAmbiguity(0);
  libSlice::ConsensusData unresults = slice->doCalculations();

  out << unresults.getConsensusBase() << " "
      << unresults.getQualityValue()  << " " 
      << slice->getNucData() << " ";

  int depth = slice->getDepth();

  int i;
  for (i = 0; i < depth; i++)
  {
    if (i) { out << ":"; }
    out << (int) slice->getQualVal(i);
  }
  
  out << " ";

  for (i = 0; i < depth; i++)
  {
    if (i) { out << ":"; }
    out << slice->getRead(i)->getId();
  }

  out << endl;
}

const string & Contig::getName() const
{
  return m_contigName;

}

