#include "SWAlign.hh"

// #define _DEBUG_VERBOSE 0
// #define _DEBUG_ASSERT 0

#include "sw_align.hh"


SWAlign::SWAlign(bool verbose)
  : ContigAlign(verbose)
{

}

SWAlign::~SWAlign()
{

}

int SWAlign::fattenLeft(Contig * contig, ContigSequence * read, int distance)
{
  initializePublicAPI(contig);

  DeltaVector delta;
  long int gindex = read->getOffset()-1;
  int i;
  unsigned int alignmentType = OPTIMAL_BACKWARD_SEARCH;
  int retval = 0;

  if (gindex < 0)
  {
    // No consensus to match
    if (m_verbose) { cerr << "No consensus to match for r" << read->getReadCount() << endl; }
    return 0;
  }

  string leftTrim = "r" + read->getLeftTrim();

  if (distance != -1)
  {
    if (m_verbose) { cerr << "Trimming Left Trim to " << distance+1 << endl; }
    reverse(leftTrim.begin(), leftTrim.end());
    leftTrim.resize(distance+1);
    reverse(leftTrim.begin(), leftTrim.end());
    alignmentType = OPTIMAL_BACKWARD_SEARCH;
  }

  long int leftend = leftTrim.size()-1;

  if (leftend == 0)
  {
    // No data to extend with
    if (m_verbose) { cerr << "No left trim for r" << read->getReadCount() << endl; }
    return 0;
  }

  if (m_verbose)
  {
    cerr << "attempting left extension of r" << read->getReadCount() 
         << " at " << gindex << " with " << leftend << "bp" << endl;

    long int leftpos = gindex-leftend;
    int shift = (leftpos > 10) ? 10 : leftpos;
    m_contig->dumpRange(leftpos-shift, gindex+10, read);

    cerr << "trim:\t";
    if (leftpos > 0)
    {
      for (i = 0; i < shift; i++) cerr << "*";
      cerr << leftTrim << endl;;
    }
    else 
    {
      cerr << leftTrim.substr(-leftpos, leftTrim.length()+leftpos) << endl;
    }
  }

  string cons = "c" + m_contig->getUngappedConsensus(0, gindex+1);
  long int consend = cons.size()-1;
  if (consend == 0)
  {
    // No consensus to match
    if (m_verbose) { cerr << "No consensus to match for r" << read->getReadCount() << endl; }
    return 0;
  }

  long int leftstart = 1;
  long int consstart = consend - leftend - 150; // allow for 150 gaps max
  if (consstart < 1) consstart = 1;

  // Compute leftend, consend
  alignSearch(leftTrim.c_str(), leftend, leftstart,
              cons.c_str(), consend, consstart,
              alignmentType);

  if ((leftend > 0 && consend > 0) &&
      (alignTarget(leftTrim.c_str(), leftstart, leftend, 
                   cons.c_str(), consstart, consend,
                   delta, FORCED_FORWARD_ALIGN)))

  {
    // alignment was sucessful!
    if (m_verbose) { cerr << "leftend=" << leftend << " consend=" << consend << endl; }

    // Trim the part that did not align (and leftTrim[0])
    leftTrim.erase(0, leftstart);

    string & consensus = m_contig->getConsensus();

    unsigned int cpos = m_contig->getGappedIndex(consstart-1); // subtract cons[0]
    unsigned int rpos = 0;

    string rightEdgeGaps;

    // Starting at an gapped position is trouble
    // Promote the edge gaps before attempting to merge the extension
    while (gindex >= 0 && consensus[gindex] == '-')
    {
      if (m_verbose) { cerr << "Promoting right edge consensus gap at " << cpos << endl; }
      rightEdgeGaps += '-';
      gindex--;
    }

    // switch to unsigned after subtracting!
    unsigned int lastConsPos = gindex;

    mergeExtension(cpos, lastConsPos, rpos, leftTrim, consensus,
                   read, delta, true, +1);

    cpos--; // Return to original gindex position
    retval = read->extendLeft(leftTrim+rightEdgeGaps);
    m_contig->dumpRange(read->getOffset()-10, cpos+10, read);
  }
  else 
  {
    if (m_verbose) { cerr << "no alignment consend: " << consend << " leftend: " << leftend << endl; }
  }

  return retval;
}

int SWAlign::fattenRight(Contig * contig, ContigSequence * read, int distance)
{
  initializePublicAPI(contig);

  DeltaVector delta;

  long int gindex = read->getRightEnd()+1;
  unsigned int alignmentType = OPTIMAL_FORWARD_SEARCH;
  int retval = 0;

  if (gindex >= m_gindex)
  {
    // No consensus to match
    if (m_verbose) { cerr << "No consensus to match for r" << read->getReadCount() << endl; }
    return 0;
  }

  string rightTrim = "r" + read->getRightTrim();
  if (distance != -1)
  {
    if (m_verbose) { cerr << "Trimming Right Trim to " << distance+1 << endl; }
    rightTrim.resize(distance+1);
    alignmentType = OPTIMAL_FORWARD_SEARCH;
  }
  long int rightend = rightTrim.size()-1;

  if (rightend == 0 || gindex >= m_gindex)
  {
    // No data to extend with
    if (m_verbose) { cerr << "No right trim for r" << read->getReadCount() << endl; }
    return 0;
  }

  if (m_verbose)
  {
    cerr << "attempting right extension of r" << read->getReadCount() 
         << " at " << gindex << " with " << rightend << "bp" << endl;

    long int leftpos = gindex-10;
    if (leftpos < 0) {leftpos = 0;}
    m_contig->dumpRange(leftpos, gindex+rightend+50, read);

    cerr << "trim:\t";
    for (int i = leftpos; i < gindex-1; i++) { cerr << "?"; }
    cerr << rightTrim << endl;
  }

  // Strip the gaps from the consensus
  string cons = "c" + m_contig->getUngappedConsensus(gindex, 
                                                     m_gindex-gindex+1);
  long int consend = cons.size()-1;
  if (consend > rightend + 150) { consend = rightend + 150; }

  if (consend == 0)
  {
    // No consensus to match
    if (m_verbose) { cerr << "No consensus to match for r" << read->getReadCount() << endl; }
    return 0;
  }

  // Compute rightend, consend
  alignSearch(rightTrim.c_str(), 1, rightend, 
              cons.c_str(), 1, consend,
              alignmentType);

  if ((rightend > 0 && consend > 0) &&
      (alignTarget(rightTrim.c_str(), 1, rightend, 
                   cons.c_str(), 1, consend,
                   delta, FORCED_FORWARD_ALIGN)))
  {
    // alignment was sucessful!
    if (m_verbose) { cerr << "rightend=" << rightend << " consend=" << consend << endl; }

    // Erase the kludge from rightTrim[0] ('r')
    rightTrim.erase(rightTrim.begin());

    // Trim the part that did not align
    rightTrim.resize(rightend);

    unsigned int cpos = gindex; 
    unsigned int rpos = 0; 

    string & consensus = m_contig->getConsensus();

    // compute gapped position of last consensus base which aligns
    unsigned int lastConsPos = gindex;
    int count = 0;

    // shift to first non-gap base (probably just gindex)
    while (consensus[lastConsPos] == '-') { lastConsPos++; }

    // Now find the exact coordinate of the last *base* which aligns
    while (count < consend)
    {
      if (consensus[lastConsPos] != '-') { count++; }
      lastConsPos++;
    }

    lastConsPos--; // Step back to exact coordinate of lastAlignedBase

    mergeExtension(cpos, lastConsPos, rpos, rightTrim, consensus, read, delta, true, +1);

    retval = read->extendRight(rightTrim);
    m_contig->dumpRange(gindex-10, read->getRightEnd()+10, read);
  }
  else
  {
    if (m_verbose) { cerr << "no alignment consend: " << consend << " rightend: " << rightend << endl; }
  }

  return retval;
}

