#include "ReadStripper.hh"
#include <cstdio>

ReadStripper::ReadStripper() : RequestManager("ReadStripper")
{
  m_log.debugL("Contructed ReadStripper");
  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_MEDIUM);

  m_readCount = 0;
  m_newIndex = 0;
  m_newGindex = 0;
}
int ReadStripper::m_renumberReads(0);

string ReadStripper::m_idstrip("");
string ReadStripper::m_idlistfilename("");
string ReadStripper::m_seqstrip("");
string ReadStripper::m_seqlistfilename("");

StringSet ReadStripper::m_stripIdSet;
StringSet ReadStripper::m_stripNameSet;


ReadStripper::~ReadStripper()
{
  m_log.debugL("Destructed ReadStripper");
}


void ReadStripper::addRequest(ConsensusRequest * request)
{
  m_requestData = request;
  printXMLHeader(request);

  m_out << "  <ReadCollection>" << endl;
}

void ReadStripper::addReadData(ReadData * readData)
{
  // The readData->getId() can change, so don't take a reference!
  const string origId = readData->getId();
  const string & seqname = readData->getSeqname();

  if (m_stripIdSet.find(origId) == m_stripIdSet.end())
  {
    // Not in the Id set, but may be in the name set
   
    if (m_stripNameSet.find(seqname) == m_stripNameSet.end())
    {
      // Not found in the Name set either, it is safe to print
      if (m_renumberReads)
      {
        char buffer[32];
        snprintf(buffer, sizeof(buffer), "%d", m_readCount);
        m_readCount++;

        m_log << "Mapping from \"" << origId 
              << "\" to \"" << buffer 
              << "\" for slice output" << endl;

        m_readMap[origId] = (string)buffer;

        // Reset the id for printing
        readData->resetId(buffer);
      }

      // Print this read (with the new id)
      readData->printXML(m_out);

      if (m_renumberReads)
      {
        // Now set it back to the original id for strip processing
        readData->resetId(origId.c_str());
      }
    }
    else
    {
      // Add this read's id to the stripIdSet
      // It could be removed from the name from the name set, but that may 
      // be more trouble than its worth.
      m_stripIdSet.insert(origId);

      m_log << "Found id \"" << origId
            << "\" with name \"" << seqname 
            << "\" to strip" << endl;
    }
  }
  else
  {
    // For consistency, add this to the name set
    m_stripNameSet.insert(seqname);

    m_log << "Found id \"" << origId 
          << "\" to strip" << endl;
  }
}



void ReadStripper::addSliceRange(long & index, long & gindex)
{
  m_out << "  </ReadCollection>" << endl;

  m_out << "  <SliceRange Index=\"" << index
        << "\" Gindex=\""           << gindex
        << "\">" << endl;

  index = gindex = 0L;
}

void ReadStripper::printSlice(SliceData * slice)
{
  char consensus = slice->getConsensus();
  long coordinate = m_newIndex;
  if (coordinate && consensus == '-') coordinate--;
  slice->resetIndex(coordinate, m_newGindex);
  slice->printXML(m_out);

  if (consensus != '-') m_newIndex++;
  m_newGindex++;
}


void ReadStripper::addSlice(SliceData * slice)
{
  unsigned short depth = slice->getDepth();
  bool first = true;
  bool changeMade = false;
  char buffer[32];

  // Store the original nucs and quals less the ones stripped
  string nucString = "";
  string qualString = "";

  // Store the original read ids less the ones stripped
  string origReadString = "";

  // Store the renumbered ids
  string newReadString = "";

  for (unsigned short i = 0; i < depth; i++)
  {
    const ReadData * rd = slice->getRead(i);
    const string & seqid = rd->getId();

    if (m_stripIdSet.find(seqid) == m_stripIdSet.end())
    {
      if (!first)
      {
        qualString += " ";
        newReadString += " ";
        origReadString += " ";
      }
      else
      {
        first = false;
      }

      snprintf(buffer, sizeof(buffer), "%d", slice->getQualVal(i));
      qualString += buffer;

      nucString += slice->getBase(i);
      origReadString += seqid;

      if (m_renumberReads)
      {
        newReadString += m_readMap[seqid];
      }
    }
    else
    {
      // This (base, qual, read) comes from a stripped read
      changeMade = true;

      m_log.pushMessageLevel(MESSAGE_LEVEL_DEBUG_HIGH);
      m_log << "Stripped data from " << rd->getSeqname() 
            << " " << rd->getId() 
            << " \"" << slice->getReadIdData()
            << "\""
            << endl;
      m_log.popMessageLevel();
    }
  }

  if (changeMade)
  {
    // Set the read string relative to the original read numbers
    // So the read information doesn't have to change in the readManager
    slice->resetReadId();
    slice->resetNuc();
    slice->resetQualVal();

    slice->addNuc(nucString.c_str());
    slice->addQualVal(qualString.c_str());
    slice->addReadId(origReadString.c_str());
  }

  // Recall the consensus against the original ids
  RequestManager::addSlice(slice);

  if (m_renumberReads)
  {
    // But print with the new ids
    slice->resetReadId();
    slice->addReadId(newReadString.c_str());
  }

  printSlice(slice);

  delete slice;
}

int ReadStripper::printResults()
{
  m_out << "  </SliceRange>" << endl;
  m_out << "</Request>"      << endl;

  return m_errorCount;
}


void ReadStripper::handleOptions(Foundation * tf)
{
  tf->getOptions()->addOptionResult("i|id=s", &m_idstrip,
                                    "Specify id to strip");

  tf->getOptions()->addOptionResult("I|idlist=s", &m_idlistfilename,
                                    "Specify file of ids to strip");

  tf->getOptions()->addOptionResult("n|seqname=s", &m_seqstrip,
                                    "Specify sequence name to strip");

  tf->getOptions()->addOptionResult("N|seqlist=s", &m_seqlistfilename,
                                    "Specify file of seqnames to strip");

  tf->getOptions()->addOptionResult("r|renumber", &m_renumberReads,
                                    "Renumber Reads after stripping");

  RequestManager::handleOptions(tf);

  if (!m_idstrip.empty())
  {
    m_stripIdSet.insert(m_idstrip);
  }

  if (!m_seqstrip.empty())
  {
    m_stripNameSet.insert(m_seqstrip);
  }

  if (!m_idlistfilename.empty())
  {
    // Add all ids from file to m_stripIdSet
    ifstream idfile(m_idlistfilename.c_str());
    string id;

    while (idfile >> id)
    {
      if (!id.empty())
      {
        m_stripIdSet.insert(id);
      }
    }
  }

  if (!m_seqlistfilename.empty())
  {
    // Add all seqnames to m_stripNameSet
    ifstream seqfile(m_seqlistfilename.c_str());
    string seq;

    while (seqfile >> seq)
    {
      if (!seq.empty())
      {
        m_stripNameSet.insert(seq);
      }
    }
  }
}
